#version 430 core
out vec4 fragColor;

in VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
} vert;

uniform sampler2D diffuseTexture;
uniform samplerCube depthCubemap;

uniform vec3 camera_world;
uniform vec3 pointLightPos;
uniform float far_plane;
uniform vec3 color;

//https://learnopengl.com/Advanced-Lighting/Shadows/Point-Shadows
float calcShadow(vec3 fragPos) {
	vec3 fragToLight = fragPos - pointLightPos; //direction vector from fragment to point light
	float currentDepth = length(fragToLight);  
	
	// include PCF to avoid shadow acne
	vec3 sampleOffsetDirections[20] = vec3[](
	vec3( 1,  1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1,  1,  1),
	vec3( 1,  1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1,  1, -1),
	vec3( 1,  1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1,  1,  0),
	vec3( 1,  0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1,  0, -1),
	vec3( 0,  1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0,  1, -1));
	
	float shadow = 0.0;
	float bias   = 0.15;
	int samples  = 20;
	float viewDistance = length(camera_world - fragPos);
	// change diskRadius based on the distance of the viewer to the fragment (shadows softer when far away and sharper when close)
	float diskRadius = (1.0 + (viewDistance / far_plane)) / 25.0;
	for(int i = 0; i < samples; ++i)
	{
		// gets uv of cubemap using direction vector (no need to normalize)
		float closestDepth = texture(depthCubemap, fragToLight + sampleOffsetDirections[i] * diskRadius).r;
		// transform from [0,1] back to [0, far_plane]
		closestDepth *= far_plane;
		// compare closest and currentDepth and check if fragment is in shadow
		if(currentDepth - bias > closestDepth)
			shadow += 1.0;
	}
	shadow /= float(samples);
	return shadow;
}

vec3 calcPointLight() {
	
	vec3 lightColor = vec3(0.3f);
	vec3 normal = normalize(vert.normal_world);
	vec3 lightDirection = normalize(pointLightPos - vert.position_world);
	vec3 viewDirection = normalize(camera_world - vert.position_world);
	//--?--vec3 reflection = reflect(-lightDirection, normal);
	
    // attenuation (fading of point light)
	float constant =  1.0f;
	float linear = 0.09f;
	float quadratic = 0.032f;
    float distance = length(pointLightPos - vert.position_world);
    float attenuation = 1.0 / (constant + linear * distance + quadratic * (distance * distance));
	
	// diffuse = light that comes from one direction (brighter if it comes squarely down)
	float diffuse = 0.6f;
	diffuse = max(dot(normal, lightDirection), 0.0f) * diffuse * attenuation;
	
	// specular = shininess
	float specular;
	if (diffuse != 0.0f) {
		vec3 halfwayDir = normalize(lightDirection + viewDirection);  
		specular = pow(max(dot(normal, halfwayDir), 0.0), 32.0) * attenuation;
	} else {
		specular = 0.0f;
	}
	
	// ambient = light with no origin (bounced off too many times) -> tunnel = very small amount
	float ambient = 0.05f;
	
	// shadow
	float shadow = calcShadow(vert.position_world);
	return lightColor * (ambient + (1.0 - shadow) * (attenuation + diffuse + specular));
}

void main() {
	vec3 texColor = texture(diffuseTexture, vert.uv).rgb;
	fragColor = vec4(color + calcPointLight() * texColor , 1);
}

